-- AGIL IMHP V3 RELAY
dofile('./interface/interface.lua');
dofile('./interface/adv.lua');
dofile('./interface/device.lua');

-- Information : Numro de Version, Nom, Interface
function device.GetInformation()
	return { 
		version = "1.2",
		name = 'AGIL IMHP-Relay V3', 
		class = 'chrono', 
		interface = { { type='serial', baudrate = '19200' }	} 
	};
end	

-- Creation et initialisation table imhp
imhp = {};

imhp.valueRSSI = -1;

-- Type de Synchro
imhp.Synchro = 
{
	NONE	= '?',
	TOP		= 'T',
	RF		= 'R',
	GPS		= 'G',
	DCF77	= 'D',
	OUT		= 'O'
};

-- Packet Header
imhp.Header = 
{
	BATTERY			= 'B',
	PULSE			= 'P',
	SYNCHRO			= 'S',
	WRITE_CONFIG	= 'W',
	READ_CONFIG		= 'R',
	DEBUG			= '?',
	RELOAD			= 'L'
};

imhp.HeaderID = 
{
	BATTERY			= 'b',
	PULSE			= 'p',
	SYNCHRO			= 's',
	WRITE_CONFIG	= 'w',
	READ_CONFIG		= 'r',
	RELOAD			= 'l'
};

imhp.Config = 
{
	ID					= 'I',
	DEBUG				= 'D',
	POWER				= 'P',
	CHANNEL				= 'C',
	DATA_INTERFACE		= 'A',
	BUZZER				= 'Z',
	INDEX_PULSE			= 'N',
	DATE				= 'd',
	DATETIME			= 't',
	TIME				= 'T',
	DELAY				= 'Y',
	CALIBRATION_N		= 'u',
	CALIBRATION_D		= 'e',
	OFFSET_RF			= 'o',
	PRINTER_BAUDRATE	= 'B',
	VERSION				= 'v',
	RSSI_MODE			= 'r',
	PRODUCT_ID			= 'H'
};

-- Calcul du ChekcSum 
imhp.ComputeCheckSum = function(packet, iStart, iEnd)
	if iStart == nil then iStart = 1 end
	if iStart < 1 then iStart = 1 end
	
	if iEnd == nil then iEnd = #packet end
	if iEnd > #packet then iEnd = #packet end

	local checkSum = 0;
	if type(packet) == 'table' then
		for i=iStart, iEnd do
			checkSum = byte.Sum(checkSum, packet[i]);
		end
	elseif type(packet) == 'string' then
		for i=iStart, iEnd do
			checkSum = byte.Sum(checkSum, string.byte(packet,i));
		end
	end
	return byte.Not(checkSum);
end

-- Ouverture
function device.OnInit(params, node)
	-- Appel OnInit Metatable
	mt_device.OnInit(params);
	
	imhp.params = params;
	imhp.node = node;
end

function imhp.alert(txt)
	app.GetAuiMessage():AddLine(txt);
end

function imhp.success(txt)
	app.GetAuiMessage():AddLineSuccess(txt);
end

function imhp.warning(txt)
	app.GetAuiMessage():AddLineWarning(txt);
end

function imhp.error(txt)
	app.GetAuiMessage():AddLineError(txt);
end

-- Fermeture
function device.OnClose()
	-- Appel OnClose Metatable
	mt_device.OnClose();
end

-- Standard Serial Event
function device.OnSerial(evt)
	if evt:GetInt() == serialNotify.RXCHAR then
		if mt_device ~= nil and mt_device.obj ~= nil then
			mt_device.obj:ReadToCircularBuffer();
			device.OnRead(mt_device.obj:GetCircularBuffer());
			mt_device.obj:ReStartEvent();
		end
	elseif evt:GetInt() == serialNotify.CONNECTION then
		imhp.success(evt:GetString());
	else
		imhp.error(evt:GetString());
	end
end

-- Lecture du Buffer Circulaire ...
function device.OnRead(cb)
	-- Test si le Buffer est vide 
	if cb:GetCount() == 0 then	return end

	-- Lecture des Packets 
	while (ReadPacket(cb)) do end
end

-- Structure des Trames IMHP : <lg> .... <checksum><cr> ou <lg> .... <checksum><rssi><cr>
-- lg 		: Longueur : 1 Byte 
-- checksum	: 2 Bytes (Hexa)
-- rssi		: / + 2 Bytes (Hexa)
function ReadPacket(cb)
	local iFind = cb:Find(asciiCode.CR);	-- Recherche CR = caractere fin de Trame 		
	if iFind == -1 then return false end 	-- On peut stopper la recherche
	
	if iFind == 0 then 
		-- Il faut traiter la cas ou CR=13 indique la longueur de la trame ...
		iFind = cb:FindChar(asciiCode.CR, 1);	-- On cherche alors la deuxime occurence
		if iFind == -1 then return false end 
	end

	-- Prise du packet 
	local packet = cb:ReadByte(iFind);

	if IsPacketOk(packet) then
		local header = string.char(packet[1]);
		if header == imhp.HeaderID.BATTERY then
			ReadPacketBattery(packet);
			MessagePacket(packet, ' (*)');
		elseif header == imhp.HeaderID.PULSE then 
			ReadPacketPulse(packet);
			MessagePacket(packet, ' (*)');
		elseif header == imhp.HeaderID.READ_CONFIG then 
			ReadPacketConfig(packet);
			MessagePacket(packet, ' (*)');
		elseif header == imhp.HeaderID.SYNCHRO then 
			ReadPacketSynchro(packet);
			MessagePacket(packet, ' (*)');
		elseif header == imhp.HeaderID.RELOAD then 
			ReadPacketPulse(packet);
			MessagePacket(packet, ' (*)');
		else
			imhp.warning("Header unknown : "..header);
			MessagePacket(packet);
		end
	else
		MessagePacket(packet);
	end

	return true;
end

-- <HEADER_ID.BATTERY><ID><PERCENT><TYPE_SYNCHRO>
function ReadPacketBattery(packet)
	local lenPacket = #packet;
	if lenPacket < 5 then return end;

	local idChrono = string.char(packet[2]);
	local percent = adv.PacketString(packet, 3, 4);
	local synchro = string.char(packet[5]);

	local infoRSSI = '';
	if synchro == 'R' then
		infoRSSI = 'RF ';
	elseif synchro == 'T' then
		infoRSSI = 'TOP ';
	end
	
	if imhp.valueRSSI ~= -1 then
		infoRSSI = infoRSSI..'rssi='..imhp.valueRSSI;
	end
	
	if lenPacket >= 9 and idChrono == '1' then
		-- Dossard au dpart ...
		local bibAck = adv.PacketString(packet, 7, 9);
		bibAck = tonumber(bibAck) or 0;
		if bibAck > 0 then
			local bibMode = string.char(packet[6]);
			imhp.bibAck = imhp.bibAck or 0;
			imhp.bibMode = imhp.bibMode or ' ';

			if bibMode == 'N' then
				-- NEXT BIB
				if imhp.bibAck ~= bibAck or imhp.bibMode ~= bibMode then 
					imhp.alert("Dossard Suivant="..bibAck);
					app.SendNotify("<bib_next>", { bib = bibAck,  passage = 0, device = 'agil_imhp', src = 'imhp' });
					imhp.bibAck = bibAck;
					imhp.bibMode = bibMode;
				end
			elseif bibMode == 'C' then
				-- CANCEL BIB
				if imhp.bibAck ~= bibAck or imhp.bibMode ~= bibMode then 
					imhp.alert("Dossard Annul="..bibAck);
					app.SendNotify("<bib_cancel>", { bib = bibAck,  passage = 0, device = 'agil_imhp', src = 'imhp' });
					imhp.bibAck = bibAck;
					imhp.bibMode = bibMode;
				end
			end
		end
	end
end

-- <HEADER_ID.PULSE><ID><INDEX (2 BYTE en BASE64)><TIME (6 BYTE en BASE64)>
-- <HEADER_ID.RELOAD><ID><INDEX (2 BYTE en BASE64)><TIME (6 BYTE en BASE64)>
function ReadPacketPulse(packet)

	local idChrono = string.char(packet[2]);
	local index64 = adv.PacketString(packet,3,4)
	local index = base64.ToInteger(index64, 2)
	local time64 = adv.PacketString(packet,5,10)
	local mstime = base64.ToInteger(time64)
	
	local origin = '9';
	local lenPacket = #packet;
	if lenPacket >= 11 then
		origin = string.char(packet[11]);
	end
	
	if lenPacket >= 15 and idChrono == '1' then
		-- Dossard au dpart ...
		local bibAck = adv.PacketString(packet, 13, 15);
		bibAck = tonumber(bibAck) or 0;
		if bibAck > 0 then
			local bibMode = string.char(packet[12]);
			imhp.bibAck = imhp.bibAck or 0;
			imhp.bibMode = imhp.bibMode or ' ';

			if bibMode == 'N' then
				-- NEXT BIB
				if imhp.bibAck ~= bibAck or imhp.bibMode ~= bibMode then 
					imhp.alert("Dossard Suivant="..bibAck);
					app.SendNotify("<bib_next>", { bib = bibAck,  passage = 0, device = 'agil_imhp', src = 'imhp' });
					imhp.bibAck = bibAck;
					imhp.bibMode = bibMode;
				end
			elseif bibMode == 'C' then
				-- CANCEL BIB
				if imhp.bibAck ~= bibAck or imhp.bibMode ~= bibMode then 
					imhp.alert("Dossard Annul="..bibAck);
					app.SendNotify("<bib_cancel>", { bib = bibAck,  passage = 0, device = 'agil_imhp', src = 'imhp' });
					imhp.bibAck = bibAck;
					imhp.bibMode = bibMode;
				end
			end
		end
	end
	
	local idPassage = tonumber(origin)-1;
	if idPassage < 0 then
		idPassage = 0;
	elseif idPassage > 4 then
		idPassage = -1;
	end
	
	AddTimePassage(mstime, idPassage, index);
end

-- <HEADER_ID.READ_CONFIG><ID><key1><value1><key2><value2>... 
function ReadPacketConfig(packet)
	local lg = #packet;
	if lg < 2 then return end;

	local ID = adv.PacketString(packet,2,2);
	if ID == 'M' then
		imhp.warning("ID="..ID..' (Chrono Master)');
	else
		imhp.warning("ID="..ID..' (Chrono Radio '..ID..')');
	end

	local i = 3;
	while i < lg do
		local key = adv.PacketString(packet,i,i);
		if key == imhp.Config.ID then
			imhp.warning("ID="..adv.PacketString(packet,i+1,i+1));
			i = i+2;
		elseif key == imhp.Config.DEBUG then
			imhp.warning("DEBUG="..adv.PacketString(packet,i+1,i+1));
			i = i+2;
		elseif key == imhp.Config.DELAY then
			imhp.warning("DELAY="..adv.PacketString(packet,i+1,i+3));
			i = i+4;
		elseif key == imhp.Config.BUZZER then
			imhp.warning("BUZZER="..adv.PacketString(packet,i+1,i+3));
			i = i+4;
		elseif key == imhp.Config.INDEX_PULSE then
			imhp.warning("INDEX_PULSE="..adv.PacketString(packet,i+1,i+4));
			i = i+5;
		elseif key == imhp.Config.DATE then
			imhp.warning("DATE="..adv.PacketString(packet,i+1,i+8));
			i = i+9;	
		elseif key == imhp.Config.DATETIME then
			imhp.warning("DATETIME="..adv.PacketString(packet,i+1,i+14));
			i = i+15;	
		elseif key == imhp.Config.TIME then
			imhp.warning("TIME="..adv.PacketString(packet,i+1,i+6));
			i = i+7;	
		elseif key == imhp.Config.POWER then
			imhp.warning("POWER="..adv.PacketString(packet,i+1,i+1));
			i = i+2;	
		elseif key == imhp.Config.CHANNEL then
			imhp.warning("CHANNEL="..adv.PacketString(packet,i+1,i+2));
			i = i+3;	
		elseif key == imhp.Config.DATA_INTERFACE then
			imhp.warning("DATA_INTERFACE="..adv.PacketString(packet,i+1,i+1));
			i = i+2;	
		elseif key == imhp.Config.RSSI_MODE then
			imhp.warning("RSSI_MODE="..adv.PacketString(packet,i+1,i+1));
			i = i+2;	
		elseif key == imhp.Config.CALIBRATION_N then
			imhp.warning("CALIBRATION_N="..adv.PacketString(packet,i+1,i+10));
			i = i+11;	
		elseif key == imhp.Config.CALIBRATION_D then
			imhp.warning("CALIBRATION_D="..adv.PacketString(packet,i+1,i+10));
			i = i+11;	
		elseif key == imhp.Config.OFFSET_RF then
			imhp.warning("OFFSET_RF="..adv.PacketString(packet,i+1,i+3));
			i = i+4;	
		elseif key == imhp.Config.PRODUCT_ID then
			imhp.warning("PRODUCT_ID="..adv.PacketString(packet,i+1,i+10));
			i = i+11;	
		elseif key == imhp.Config.VERSION then
			imhp.warning("VERSION="..adv.PacketString(packet,i+1,i+4));
			i = i+5;	
		elseif key == imhp.Config.PRINTER_BAUDRATE then
			local baudrate = adv.PacketString(packet,i+1,i+1);
			i = i+2;	
			if baudrate == imhp.Printer.BAUDRATE_1200 then
				imhp.warning("PRINTER=1200");
			elseif baudrate == imhp.Printer.BAUDRATE_2400 then
				imhp.warning("PRINTER=2400");
			elseif baudrate == imhp.Printer.BAUDRATE_4800 then
				imhp.warning("PRINTER=4800");
			elseif baudrate == imhp.Printer.BAUDRATE_9600 then
				imhp.warning("PRINTER=9600");
			else
				imhp.warning("PRINTER=???");
			end
		else
			imhp.warning("KEY="..key.." UNKNOWN ???");
			i = lg;
			break;
		end
	end
end

-- <PACKET_HEADER_ID_SYNCHRO><id><SYNCHRO_TOP><HHMMSS><origin>
-- <PACKET_HEADER_ID_SYNCHRO><id><SYNCHRO_RF>
function ReadPacketSynchro(packet)

	local lg = #packet;
	if lg < 3 then return end;

	local idChrono = adv.PacketString(packet, 2, 2);
	local typeSynchro = adv.PacketString(packet, 3, 3);
	
	if typeSynchro == imhp.Synchro.TOP then
		
		local strHHMMSS = adv.PacketString(packet, 4, 9);
		imhp.warning("SYNCHRO TOP "..strHHMMSS);
		local mstime = adv.HHMMSS_To_MS(strHHMMSS);
		local origin = adv.PacketString(packet, 10, 10);
				
		local countChrono = #imhp.chrono;
		for i=1,countChrono do
			if imhp.chrono[i].id == idChrono then
				imhp.chrono[i].tickcount = app.GetTickCount();

				if origin == '2' then
					AddTimePassage(mstime, imhp.chrono[i].input2, -1);
				else
					AddTimePassage(mstime, imhp.chrono[i].input1, -1);
				end
				return;
			end
		end
	elseif typeSynchro == imhp.Synchro.RF then
		imhp.warning("SYNCHRO RF "..idChrono);
	end 
end

function IsPacketOk(packet)
	
	local valueRSSI = imhp.valueRSSI;

	if IsPacketKeyboard(packet) then
		MessagePacket(packet, " (Keyboard)");
	end

	local lg = #packet;
	if lg < 4 then return false end
	if packet[lg] ~= asciiCode.CR then return false end

	local lg1 = packet[1];
	if packet[lg-3] == string.byte('/') then
		-- Information RSSI prsente dans la trame ...
		if lg > 7 then
			local rssiHexa = string.char(packet[lg-2])..string.char(packet[lg-1]);
			valueRSSI = hexa.ToInteger(rssiHexa); 
			-- Suppression des 3 Bytes RSSI
			table.remove(packet, #packet-1);
			table.remove(packet, #packet-1);
			table.remove(packet, #packet-1);
			lg = #packet;
		end
	end
	
	-- Test cohrence sur la longueur 
	if lg ~= lg1 + 1 then return false end

	-- Test checkSum
	local checksum1 = imhp.ComputeCheckSum(packet, 2, lg - 3);
	local checksumString = string.char(packet[lg-2])..string.char(packet[lg-1]);
	local checksum2 = hexa.ToInteger(checksumString); 
	if checksum1 ~= checksum2 then return false; end
	
	-- Longueur Ok, CheckSum Ok => la trame est correcte ...
	table.remove(packet,1); -- Suppression Byte Longueur
	table.remove(packet, #packet);	-- Suppression CR
	table.remove(packet, #packet);	-- Suppression CheckSum deuxieme caractere
	table.remove(packet, #packet);	-- Suppression CheckSum premier caractere

	imhp.valueRSSI = valueRSSI;
	return true;
end

function IsPacketKeyboard(packet)
	if #packet < 4 then return false end
	if (packet[1] ~= string.byte('[') or packet[4] ~= string.byte(']')) then return false end

	-- Remplacement de la sequence [xx] par la vraie longueur sur 1 Byte 
	local lgHexa = string.char(packet[2])..string.char(packet[3]);
	local lg = hexa.ToInteger(lgHexa); 
	table.remove(packet, 1);
	table.remove(packet, 1);
	table.remove(packet, 1);
	packet[1] = lg+2;

	-- Ajout du CheckSum ...
	lg = #packet;
	local checkSum = imhp.ComputeCheckSum(packet, 2, lg-1);
	local strCheckSum = hexa.ToString(checkSum, 2);
	table.insert(packet, lg, string.byte(strCheckSum,1));
	table.insert(packet, lg+1, string.byte(strCheckSum,2));

	return true;
end

-- Envoi du packet dans la fentre d'information
function MessagePacket(packet, msg)
	local strPacket = '';
	for i=1,#packet do
		if packet[i] >= 32 then
			strPacket = strPacket..string.char(packet[i]);
		end
	end

	if msg == nil then
		imhp.alert(strPacket);
	else
		imhp.alert(strPacket..msg);
	end
end

function AddTimePassage(chrono, passage, index)
	app.SendNotify("<passage_add>", 
		{ time = chrono,  passage = passage, device = 'agil_imhp', log=index }
	);
end



